@@ -25,6 +25,17 @@ static geoipv6_t IPV6_NULL;
 
 #if defined(_WIN32)
 #include <io.h>
+
+#ifdef _MSC_VER
+#  if _MSC_VER < 1900 // VS 2015 supports snprintf
+#    define snprintf _snprintf
+#  endif
+#  if _MSC_VER >= 1400 // VS 2005+ deprecates fileno, lseek and read
+#    define fileno _fileno
+#    define read _read
+#    define lseek _lseek
+#  endif
+#endif
 #else
 #include <unistd.h>
 #include <netdb.h>
@@ -629,10 +640,8 @@ static const char * _GeoIP_inet_ntop(int af, const void *src, char *dst,
 {
     return inet_ntop(af, src, dst, cnt);
 }
-
 #endif /* defined(_WIN32) */
 
-
 int __GEOIP_V6_IS_NULL(geoipv6_t v6)
 {
     int i;
@@ -918,8 +927,34 @@ static int _database_has_content( int database_type )
             && database_type != GEOIP_REGION_EDITION_REV1) ? 1 : 0;
 }
 
-static
-void _setup_segments(GeoIP * gi)
+
+static ssize_t get_index_size(GeoIP * gi, struct stat *buf)
+{
+    ssize_t index_size;
+    unsigned int segment;
+
+    if (!_database_has_content(gi->databaseType)) {
+        return buf->st_size;
+    }
+
+    segment = gi->databaseSegments[0];
+    index_size = segment * (ssize_t)gi->record_length * 2;
+
+    /* check for overflow in multiplication */
+    if (segment != 0
+        && index_size / segment != (ssize_t)gi->record_length * 2) {
+        return -1;
+    }
+
+    /* Index size should never exceed the size of the file */
+    if (index_size > buf->st_size) {
+        return -1;
+    }
+
+    return index_size;
+}
+
+static void _setup_segments(GeoIP * gi)
 {
     int i, j, segment_record_length;
     unsigned char delim[3];
@@ -1036,10 +1071,10 @@ void _setup_segments(GeoIP * gi)
 }
 
 static
-int _check_mtime(GeoIP *gi)
+void _check_mtime(GeoIP *gi)
 {
     struct stat buf;
-    unsigned int idx_size;
+    ssize_t idx_size;
 
 #if !defined(_WIN32)
     struct timeval t;
@@ -1053,9 +1088,15 @@ int _check_mtime(GeoIP *gi)
 #if !defined(_WIN32)
         /* stat only has second granularity, so don't
          * call it more than once a second */
-        gettimeofday(&t, NULL);
+        if (0 != gettimeofday(&t, NULL)) {
+            DEBUG_MSGF(gi->flags,
+                       "Error calling gettimeofday: %s\n",
+                       strerror(errno));
+            return;
+        }
+
         if (t.tv_sec == gi->last_mtime_check) {
-            return 0;
+            return;
         }
         gi->last_mtime_check = t.tv_sec;
 
@@ -1066,7 +1107,7 @@ int _check_mtime(GeoIP *gi)
         GetSystemTimeAsFileTime(&ft);
         t = FILETIME_TO_USEC(ft) / 1000 / 1000;
         if (t == gi->last_mtime_check) {
-            return 0;
+            return;
         }
         gi->last_mtime_check = t;
 
@@ -1096,7 +1137,7 @@ int _check_mtime(GeoIP *gi)
                             DEBUG_MSGF(gi->flags,
                                        "Out of memory when reloading %s\n",
                                        gi->file_path);
-                            return -1;
+                            return;
                         }
                     }
                 }
@@ -1107,7 +1148,7 @@ int _check_mtime(GeoIP *gi)
                     DEBUG_MSGF(gi->flags,
                                "Error Opening file %s when reloading\n",
                                gi->file_path);
-                    return -1;
+                    return;
                 }
                 gi->mtime = buf.st_mtime;
                 gi->size = buf.st_size;
@@ -1117,7 +1158,7 @@ int _check_mtime(GeoIP *gi)
                     DEBUG_MSGF(gi->flags,
                                "GEOIP_MMAP_CACHE is not supported on WIN32\n");
                     gi->cache = 0;
-                    return -1;
+                    return;
 #else
                     gi->cache =
                         mmap(NULL, buf.st_size, PROT_READ, MAP_PRIVATE, fileno(
@@ -1129,7 +1170,7 @@ int _check_mtime(GeoIP *gi)
                                    gi->file_path);
 
                         gi->cache = NULL;
-                        return -1;
+                        return;
                     }
 #endif
                 } else if (gi->flags & GEOIP_MEMORY_CACHE) {
@@ -1138,7 +1179,7 @@ int _check_mtime(GeoIP *gi)
                         DEBUG_MSGF(gi->flags,
                                    "Error reading file %s when reloading\n",
                                    gi->file_path);
-                        return -1;
+                        return;
                     }
                 }
 
@@ -1150,21 +1191,14 @@ int _check_mtime(GeoIP *gi)
                 if (gi->databaseSegments == NULL) {
                     DEBUG_MSGF(gi->flags, "Error reading file %s -- corrupt\n",
                                gi->file_path);
-                    return -1;
+                    return;
                 }
 
-                /* make sure the index is <= file size
-                 * This test makes sense for all modes - not
-                 * only index
-                 */
-                idx_size =
-                    _database_has_content(gi->databaseType) ? gi->
-                    databaseSegments[0
-                    ] * (long)gi->record_length * 2 :  buf.st_size;
-                if (idx_size > buf.st_size) {
+                idx_size = get_index_size(gi, &buf);
+                if (idx_size < 0) {
                     DEBUG_MSGF(gi->flags, "Error file %s -- corrupt\n",
                                gi->file_path);
-                    return -1;
+                    return;
                 }
 
                 if (gi->flags & GEOIP_INDEX_CACHE) {
@@ -1172,18 +1206,19 @@ int _check_mtime(GeoIP *gi)
                         gi->index_cache, sizeof(unsigned char) * idx_size );
                     if (gi->index_cache != NULL) {
                         if (pread(fileno(gi->GeoIPDatabase), gi->index_cache,
-                                  idx_size, 0 ) != (ssize_t)idx_size) {
-                            DEBUG_MSGF(gi->flags,
-                                       "Error reading file %s where reloading\n",
-                                       gi->file_path);
-                            return -1;
+                                  idx_size, 0 ) != idx_size) {
+                            DEBUG_MSGF(
+                                gi->flags,
+                                "Error reading file %s where reloading\n",
+                                gi->file_path);
+                            return;
                         }
                     }
                 }
             }
         }
     }
-    return 0;
+    return;
 }
 
 #define ADDR_STR_LEN (8 * 4 + 7 + 1)
@@ -1264,9 +1299,10 @@ unsigned int _GeoIP_seek_record_v6_gl(GeoIP *gi, geoipv6_t ipnum,
 
     /* shouldn't reach here */
     _GeoIP_inet_ntop(AF_INET6, &ipnum.s6_addr[0], paddr, ADDR_STR_LEN);
-    DEBUG_MSGF(gi->flags,
-               "Error Traversing Database for ipnum = %s - Perhaps database is corrupt?\n",
-               paddr);
+    DEBUG_MSGF(
+        gi->flags,
+        "Error Traversing Database for ipnum = %s - Perhaps database is corrupt?\n",
+        paddr);
     return 0;
 }
 
@@ -1352,9 +1388,10 @@ unsigned int _GeoIP_seek_record_gl(GeoIP *gi, unsigned long ipnum,
         offset = x;
     }
     /* shouldn't reach here */
-    DEBUG_MSGF(gi->flags,
-               "Error Traversing Database for ipnum = %lu - Perhaps database is corrupt?\n",
-               ipnum);
+    DEBUG_MSGF(
+        gi->flags,
+        "Error Traversing Database for ipnum = %lu - Perhaps database is corrupt?\n",
+        ipnum);
     return 0;
 }
 
@@ -1442,7 +1479,7 @@ GeoIP * GeoIP_new(int flags)
 GeoIP * GeoIP_open(const char * filename, int flags)
 {
     struct stat buf;
-    unsigned int idx_size;
+    ssize_t idx_size;
     GeoIP * gi;
     size_t len;
 
@@ -1520,12 +1557,9 @@ GeoIP * GeoIP_open(const char * filename, int flags)
     gi->ext_flags = 1U << GEOIP_TEREDO_BIT;
     _setup_segments(gi);
 
-    idx_size =
-        _database_has_content(gi->databaseType) ? gi->databaseSegments[0] *
-        (long)gi->record_length * 2 :  buf.st_size;
+    idx_size = get_index_size(gi, &buf);
 
-    /* make sure the index is <= file size */
-    if (idx_size > buf.st_size) {
+    if (idx_size < 0) {
         DEBUG_MSGF(gi->flags, "Error file %s -- corrupt\n", gi->file_path);
         if (flags & GEOIP_MEMORY_CACHE) {
             free(gi->cache);
@@ -1558,6 +1592,9 @@ GeoIP * GeoIP_open(const char * filename, int flags)
     } else {
         gi->index_cache = NULL;
     }
+
+    gi->last_mtime_check = 0;
+
     return gi;
 }
 
@@ -1598,7 +1635,7 @@ const char *GeoIP_country_code_by_name_v6_gl(GeoIP * gi, const char *name,
 {
     int country_id;
     country_id = GeoIP_id_by_name_v6_gl(gi, name, gl);
-    return (country_id > 0) ? GeoIP_country_code[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code_by_name_gl(GeoIP * gi, const char *name,
@@ -1606,7 +1643,7 @@ const char *GeoIP_country_code_by_name_gl(GeoIP * gi, const char *name,
 {
     int country_id;
     country_id = GeoIP_id_by_name_gl(gi, name, gl);
-    return (country_id > 0) ? GeoIP_country_code[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code3_by_name_v6_gl(GeoIP * gi, const char *name,
@@ -1614,7 +1651,7 @@ const char *GeoIP_country_code3_by_name_v6_gl(GeoIP * gi, const char *name,
 {
     int country_id;
     country_id = GeoIP_id_by_name_v6_gl(gi, name, gl);
-    return (country_id > 0) ? GeoIP_country_code3[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code3_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code3_by_name_gl(GeoIP * gi, const char *name,
@@ -1622,7 +1659,7 @@ const char *GeoIP_country_code3_by_name_gl(GeoIP * gi, const char *name,
 {
     int country_id;
     country_id = GeoIP_id_by_name_gl(gi, name, gl);
-    return (country_id > 0) ? GeoIP_country_code3[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code3_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_name_by_name_v6_gl(GeoIP * gi, const char *name,
@@ -1769,7 +1806,7 @@ const char *GeoIP_country_code_by_addr_v6_gl(GeoIP * gi, const char *addr,
 {
     int country_id;
     country_id = GeoIP_id_by_addr_v6_gl(gi, addr, gl);
-    return (country_id > 0) ? GeoIP_country_code[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code_by_id(country_id) : NULL;
 }
 
 
@@ -1778,14 +1815,14 @@ const char *GeoIP_country_code_by_addr_gl(GeoIP * gi, const char *addr,
 {
     int country_id;
     country_id = GeoIP_id_by_addr_gl(gi, addr, gl);
-    return (country_id > 0) ? GeoIP_country_code[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code_by_id(country_id) : NULL;
 }
 const char *GeoIP_country_code3_by_addr_v6_gl(GeoIP * gi, const char *addr,
                                               GeoIPLookup * gl)
 {
     int country_id;
     country_id = GeoIP_id_by_addr_v6_gl(gi, addr, gl);
-    return (country_id > 0) ? GeoIP_country_code3[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code3_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code3_by_addr_gl(GeoIP * gi, const char *addr,
@@ -1793,7 +1830,7 @@ const char *GeoIP_country_code3_by_addr_gl(GeoIP * gi, const char *addr,
 {
     int country_id;
     country_id = GeoIP_id_by_addr_gl(gi, addr, gl);
-    return (country_id > 0) ? GeoIP_country_code3[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code3_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_name_by_addr_v6_gl(GeoIP * gi, const char *addr,
@@ -1833,7 +1870,7 @@ const char *GeoIP_country_code_by_ipnum_gl(GeoIP * gi, unsigned long ipnum,
 {
     int country_id;
     country_id = GeoIP_id_by_ipnum_gl(gi, ipnum, gl);
-    return (country_id > 0) ? GeoIP_country_code[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code_by_ipnum_v6_gl(GeoIP * gi, geoipv6_t ipnum,
@@ -1841,7 +1878,7 @@ const char *GeoIP_country_code_by_ipnum_v6_gl(GeoIP * gi, geoipv6_t ipnum,
 {
     int country_id;
     country_id = GeoIP_id_by_ipnum_v6_gl(gi, ipnum, gl);
-    return (country_id > 0) ? GeoIP_country_code[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code3_by_ipnum_gl(GeoIP * gi, unsigned long ipnum,
@@ -1849,7 +1886,7 @@ const char *GeoIP_country_code3_by_ipnum_gl(GeoIP * gi, unsigned long ipnum,
 {
     int country_id;
     country_id = GeoIP_id_by_ipnum_gl(gi, ipnum, gl);
-    return (country_id > 0) ? GeoIP_country_code3[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code3_by_id(country_id) : NULL;
 }
 
 const char *GeoIP_country_code3_by_ipnum_v6_gl(GeoIP * gi, geoipv6_t ipnum,
@@ -1857,14 +1894,13 @@ const char *GeoIP_country_code3_by_ipnum_v6_gl(GeoIP * gi, geoipv6_t ipnum,
 {
     int country_id;
     country_id = GeoIP_id_by_ipnum_v6_gl(gi, ipnum, gl);
-    return (country_id > 0) ? GeoIP_country_code3[country_id] : NULL;
+    return (country_id > 0) ? GeoIP_code3_by_id(country_id) : NULL;
 }
 
 int GeoIP_country_id_by_addr_v6_gl(GeoIP * gi, const char *addr,
                                    GeoIPLookup * gl)
 {
-    GeoIPLookup n;
-    return GeoIP_id_by_addr_v6_gl(gi, addr, &n);
+    return GeoIP_id_by_addr_v6_gl(gi, addr, gl);
 }
 
 int GeoIP_country_id_by_addr_gl(GeoIP * gi, const char *addr, GeoIPLookup * gl)
@@ -2040,7 +2076,10 @@ void GeoIP_assign_region_by_inetaddr_gl(GeoIP * gi, unsigned long inetaddr,
             region->region[0] = (char)((seek_region - 1000) / 26 + 65);
             region->region[1] = (char)((seek_region - 1000) % 26 + 65);
         } else {
-            memcpy(region->country_code, GeoIP_country_code[seek_region], 2);
+            const char *code = GeoIP_code_by_id(seek_region);
+            if (code != NULL) {
+                memcpy(region->country_code, code, 2);
+            }
         }
     } else if (gi->databaseType == GEOIP_REGION_EDITION_REV1) {
         /* Region Edition, post June 2003 */
@@ -2061,12 +2100,13 @@ void GeoIP_assign_region_by_inetaddr_gl(GeoIP * gi, unsigned long inetaddr,
             region->region[0] = (char)((seek_region - CANADA_OFFSET) / 26 + 65);
             region->region[1] = (char)((seek_region - CANADA_OFFSET) % 26 + 65);
         } else {
-            /* Not US or Canada ( world countries country code is always cc_id * 360 ) */
-            /* coverity[dont_call] */
-            memcpy(
-                region->country_code,
-                GeoIP_country_code[(seek_region -
-                                    WORLD_OFFSET) / FIPS_RANGE], 2);
+            /* Not US or Canada ( cc_id is always cc_id * FIPS_RANGE ) */
+            const char *code = GeoIP_code_by_id(
+                (seek_region - WORLD_OFFSET) / FIPS_RANGE);
+            if (code != NULL) {
+                /* coverity[dont_call] */
+                memcpy(region->country_code, code, 2);
+            }
         }
     }
 }
@@ -2092,8 +2132,11 @@ void GeoIP_assign_region_by_inetaddr_v6_gl(GeoIP * gi, geoipv6_t inetaddr,
             region->region[0] = (char)((seek_region - 1000) / 26 + 65);
             region->region[1] = (char)((seek_region - 1000) % 26 + 65);
         } else {
-            /* coverity[dont_call] */
-            memcpy(region->country_code, GeoIP_country_code[seek_region], 2);
+            const char *code = GeoIP_code_by_id(seek_region);
+            if (code != NULL) {
+                /* coverity[dont_call] */
+                memcpy(region->country_code, code, 2);
+            }
         }
     } else if (gi->databaseType == GEOIP_REGION_EDITION_REV1) {
         /* Region Edition, post June 2003 */
@@ -2115,11 +2158,12 @@ void GeoIP_assign_region_by_inetaddr_v6_gl(GeoIP * gi, geoipv6_t inetaddr,
             region->region[1] = (char)((seek_region - CANADA_OFFSET) % 26 + 65);
         } else {
             /* Not US or Canada ( cc_id is always cc_id * FIPS_RANGE ) */
-            /* coverity[dont_call] */
-            memcpy(
-                region->country_code,
-                GeoIP_country_code[(seek_region -
-                                    WORLD_OFFSET) / FIPS_RANGE], 2);
+            const char *code = GeoIP_code_by_id(
+                (seek_region - WORLD_OFFSET) / FIPS_RANGE);
+            if (code != NULL) {
+                /* coverity[dont_call] */
+                memcpy(region->country_code, code, 2);
+            }
         }
     }
 }
@@ -2261,7 +2305,7 @@ void GeoIPRegion_delete(GeoIPRegion *gir)
 static
 char *_get_name_gl(GeoIP * gi, unsigned long ipnum, GeoIPLookup * gl)
 {
-    int seek_org;
+    unsigned int seek_org;
     char buf[MAX_ORG_RECORD_LENGTH];
     char * org_buf, * buf_pointer;
     int record_pointer;
@@ -2323,7 +2367,7 @@ char *_get_name_gl(GeoIP * gi, unsigned long ipnum, GeoIPLookup * gl)
 static
 char *_get_name_v6_gl(GeoIP * gi, geoipv6_t ipnum, GeoIPLookup * gl)
 {
-    int seek_org;
+    unsigned int seek_org;
     char buf[MAX_ORG_RECORD_LENGTH + 1];
     char * org_buf, * buf_pointer;
     int record_pointer;
@@ -2415,7 +2459,7 @@ char **GeoIP_range_by_ip_gl(GeoIP * gi, const char *addr, GeoIPLookup * gl)
     unsigned long right_seek;
     unsigned long mask;
     int orig_netmask;
-    int target_value;
+    unsigned int target_value;
     char **ret;
     GeoIPLookup t;
 
@@ -2654,4 +2698,3 @@ int GeoIP_cleanup(void)
 
     return result;
 }
-
